import re
from pathlib import Path
from typing import Dict, List, Mapping, Optional, Union

import pandas


def read_map(contents: Union[str, Path]) -> Dict[str, str]:
	"""
		Parses the contents of any file used to specify a key-value pair. This is used when defining palettes and gene aliases.
	Parameters
	----------
	contents: Union[str,Path]
		Either a path to the file or the contents of the file.

	Returns
	-------
	Dict[str,str]
	"""
	lines = readfile(contents)

	filemap = dict()
	for line in lines:
		fields = line.split('\t')
		try:
			# We only care about the first two values in the line. Anything after that is ignored. May be used to add notes ot the file.
			# It's fine if the first row contains headers. They shouldn't interfere with any of the genotype/genes/etc.
			left, right, *_ = fields
		except ValueError:
			# Not enough values in line.
			continue
		left = left.strip()
		right = right.strip()
		if left and right:
			filemap[left] = right

	return filemap


def readfile(contents: Union[str, Path]) -> List[str]:
	"""
		Takes a file path or the contents of a file and attemts to split it into lines.
	Parameters
	----------
	contents: Union[str, Path]

	Returns
	-------
	List[str]
		The lines in the file.

	"""
	if not isinstance(contents, str):
		contents = contents.read_text()
	lines = contents.split('\n')  # Use this instead of splitlines sine splitlines sometimes doesn't work right.
	# Remove any whitespace to the left and right of the contents in the line. This should avaid issues with extra whitespace padding on the lines.
	clean_lines = [l.strip() for l in lines]
	# Remove empty lines
	nonempty_lines = [l for l in clean_lines if l]
	return nonempty_lines


def parse_known_genotypes(known_genotypes: Union[str, Path]) -> List[List[str]]:
	lines = readfile(known_genotypes)

	genotypes = list()
	for line in lines:
		trajectories = line.split(',')
		trajectories = [i.strip() for i in trajectories]
		genotypes.append(trajectories)
	return genotypes


def _get_column_values(table: pandas.DataFrame, column: str) -> List[str]:
	if column not in table.columns:
		if column.title() in table.columns:
			column = column.title()
		else:
			return []

	column_values = table[column].tolist()
	column_values = [(i if isinstance(i, str) else "") for i in column_values]
	return column_values


def _clean_gene_label(label: str) -> str:
	if not isinstance(label, str):
		return ""
	gene_separator = "/>"
	if gene_separator in label:
		sublabels = [_clean_gene_label(i) for i in label.split(gene_separator)]
		clean_label = "|".join(sublabels)
	else:
		pattern = "[a-zA-Z_0-9\s]+"
		match = re.match(pattern, label)
		if match:
			clean_label = match.group(0)
		else:
			clean_label = ""
	return clean_label


def _clean_annotation_label(label: str) -> str:
	"""
		Removes extra stuff included in the annotation form breseq.
	"""
	if not isinstance(label, str): return ""
	if 'intergenic' in label:
		result = 'intergenic'
	elif 'del' in label:
		result = 'del'
	else:
		result = label.split('(')[0]
		if result == "":
			result = label
	return result


def _extract_value(row: Mapping[str, str], column: str) -> Optional[str]:
	"""
		Extracts a value from a Mapping (usually a pandas.Series object) using a few variations of the input column.
		For example, the column containing gene labels could be names 'Gene' or 'gene'. pandas.Series objects save null values as math.nan, a float.
	"""
	if column in row:
		value = row[column]
	elif column.lower() in row:
		value = row[column.lower()]
	elif column.title() in row:
		value = row[column.title()]
	else:
		value = None
	if isinstance(value, float):
		value = None
	return value


def extract_annotations(info: pandas.DataFrame, alias_filename: Optional[Path] = None) -> Dict[str, str]:
	if alias_filename:
		alias_map = read_map(alias_filename)
	else:
		alias_map = {}
	gene_column = 'gene'
	annotation_column = 'annotation'

	trajectory_annotations = dict()
	for trajectory_label, row in info.iterrows():
		gene_value = _extract_value(row, gene_column)
		annotation_value = _extract_value(row, annotation_column)
		gene = _clean_gene_label(gene_value)
		gene = alias_map.get(gene, gene)
		annotation = _clean_annotation_label(annotation_value)
		value = " ".join(i for i in [gene, annotation] if i)
		trajectory_annotations[trajectory_label] = value
	return trajectory_annotations


def parse_genotype_annotations(genotype_members: Mapping[str, Union[str, List[str]]], info: pandas.DataFrame,
		alias_filename: Optional[Path] = None) -> Dict[
	str, List[str]]:
	trajectory_annotations = extract_annotations(info, alias_filename)

	genotype_annotations = dict()
	for genotype_label, members in genotype_members.items():
		if isinstance(members, str):
			members = members.split('|')
		member_values: List[Optional[str]] = [trajectory_annotations.get(i) for i in members]
		# Remove missing annotations
		member_values: List[str] = [i for i in member_values if i]
		genotype_annotations[genotype_label] = member_values
	return genotype_annotations



import itertools
from typing import Any, List, Tuple

import pandas
from loguru import logger
from scipy.cluster import hierarchy
from scipy.spatial import distance

try:
	from muller.clustering.metrics.pairwise_calculation_cache import PairwiseCalculationCache
except ModuleNotFoundError:
	from ..metrics.pairwise_calculation_cache import PairwiseCalculationCache


def hierarchical_method(pair_array: PairwiseCalculationCache, similarity_cutoff: float, cluster_method: str = 'distance',
		starting_genotypes: List[List[str]] = None) -> Tuple[List[List[str]], Any]:
	"""

	Parameters
	----------
	pair_array
	similarity_cutoff
	cluster_method: {'distance;, 'monocrit'}; default 'monocrit'
		The method to cluster leafs by.
	starting_genotypes: List[List[str]]
		Each element should be a list of trajectories known to be in the same genotype.
	Returns
	-------

	"""
	logger.info(f"similarity_cutoff: {similarity_cutoff}")
	logger.info(f"cluster_method: {cluster_method}")

	# If known genotypes are given, modify the pair_array so that they will be grouped together.
	if starting_genotypes:
		for genotype in starting_genotypes:
			combinations = itertools.combinations(genotype, 2)
			values = dict()
			for a, b in combinations:
				values[a, b] = 0
				values[b, a] = 0
			pair_array = pair_array.update(values)

	squaremap = pair_array.squareform()
	condensed_squaremap = distance.squareform(squaremap.values)
	method = 'complete'
	Z = hierarchy.linkage(condensed_squaremap, method = method, optimal_ordering = True)
	inconsistent = hierarchy.inconsistent(Z, 10)
	MR = hierarchy.maxRstat(Z, inconsistent, 1)

	if cluster_method == 'distance':
		maximum_distance = max(pair_array.pairwise_values.values())
		distances = pandas.Series([i for i in pair_array.pairwise_values.values() if (i != maximum_distance and i > 0)])

		similarity_cutoff = distances.quantile(similarity_cutoff)

		logger.info(f"Using Hierarchical Clustering with similarity cutoff {similarity_cutoff}")

		clusters = hierarchy.fcluster(Z, t = similarity_cutoff, criterion = 'distance')
	elif cluster_method == 'monocrit':
		clusters = hierarchy.fcluster(Z, t = similarity_cutoff, criterion = 'monocrit', monocrit = MR)
	elif cluster_method == 'inconsistent':
		clusters = hierarchy.fcluster(Z, t = similarity_cutoff, criterion = 'inconsistent')
	else:
		message = f"'{cluster_method}' is not a currently supported clustering method when using hierarchical clustering."
		logger.error(message)
		raise ValueError(message)
	cluster_map = {i: list() for i in clusters}
	for i, j in zip(clusters, squaremap.index):
		cluster_map[i].append(j)
	return list(cluster_map.values()), Z



import argparse
import itertools
import math
from pathlib import Path
from typing import List, Optional, Union

try:
	from muller import dataio
except ModuleNotFoundError:
	from . import dataio

from dataclasses import dataclass, fields

__VERSION__ = "0.4.10"


# For convienience. Helps with autocomplete.
@dataclass
class ProgramOptions(argparse.Namespace):
	filename: Path
	output_folder: Path
	sheetname: str = "Sheet1"
	fixed_breakpoint: Optional[float] = None
	detection_breakpoint: float = 0.03
	significant_breakpoint: float = 0.15
	mode: bool = False
	frequencies: List[float] = 0.10
	similarity_breakpoint: float = 0.05
	difference_breakpoint: float = 0.10
	is_genotype: bool = False
	use_filter: bool = True
	annotate_all: bool = False
	save_pvalue: bool = True
	use_strict_filter: bool = False
	method: str = 'matlab'
	metric: str = "similarity"
	known_genotypes: Optional[Path] = None

	def show(self):
		for field in fields(self):
			print(field)


ACCEPTED_METHODS = ["matlab", "hierarchy", "twostep"]


def parse_workflow_options(program_options: ProgramOptions) -> ProgramOptions:
	"""
		Generates values for each of the program parameters from the given parameters on the command line.
	Parameters
	----------
	program_options

	Returns
	-------

	"""

	if program_options.fixed_breakpoint is None:
		program_options.fixed_breakpoint = 1 - program_options.detection_breakpoint
	if program_options.additive_cutoff is None:
		program_options.additive_cutoff = program_options.detection_breakpoint
	if program_options.subtractive_cutoff is None:
		program_options.subtractive_cutoff = program_options.detection_breakpoint

	# TODO read the additional input files here rather than at the end of the analysis.
	if program_options.known_genotypes:
		program_options.known_genotypes = Path(program_options.known_genotypes)
		starting_genotypes = dataio.parse_known_genotypes(program_options.known_genotypes)
	else:
		starting_genotypes = None

	program_options.starting_genotypes = starting_genotypes
	cluster_method = program_options.method
	if cluster_method not in ACCEPTED_METHODS:
		message = f"{cluster_method} is not a valid option for the --method option. Expected one of {ACCEPTED_METHODS}"
		raise ValueError(message)
	return program_options


def _parse_frequency_option(frequency: Union[str, List[float]]) -> List[float]:
	if isinstance(frequency, str):
		if ',' in frequency:
			frequencies = list(map(float, frequency.split(',')))
		else:
			frequencies = float(frequency)
	elif isinstance(frequency, float):
		frequencies = frequency
	else:
		frequencies = [float(i) for i in frequency]
	if isinstance(frequencies, float):
		frequencies = [math.fsum(itertools.repeat(frequencies, i)) for i in range(int(1 / frequencies) + 1)]
		frequencies = [round(i, 2) for i in frequencies]
	if 0 not in frequencies:
		frequencies = [0] + frequencies
	frequencies = sorted(frequencies, reverse = True)
	return frequencies


class FrequencyParser(argparse.Action):
	def __call__(self, parser, namespace, values, option_string = None):
		if not values:
			parsed_frequencies = self.default
		else:
			parsed_frequencies = _parse_frequency_option(values)
		setattr(namespace, self.dest, parsed_frequencies)


class FixedBreakpointParser(argparse.Action):
	def __call__(self, parser, namespace, values, option_string = None):
		detected = namespace.detection_breakpoint
		if values == "1":
			values = 1 - detected
		else:
			values = float(values)
		setattr(namespace, self.dest, values)


def create_parser() -> argparse.ArgumentParser:
	parser = argparse.ArgumentParser(
		description = "Generates muller diagrams based on a set of mutational trajectories.",
		formatter_class = argparse.ArgumentDefaultsHelpFormatter
	)
	##############################################################################################################################################
	# --------------------------------------------------------- Required Parameters --------------------------------------------------------------
	##############################################################################################################################################
	parser.add_argument(
		"-v", "--version",
		action = 'version',
		version = f"%(prog)s {__VERSION__}"
	)
	parser.add_argument(
		'-i', '--input',
		help = "The table of trajectories to cluster.",
		action = 'store',
		dest = 'filename',
		type = Path,
		required = True
	)
	parser.add_argument(
		'-o', '--output',
		help = "The folder to save the files to.",
		action = 'store',
		dest = 'output_folder',
		type = Path,
		required = True
	)
	parser.add_argument(
		"--name",
		help = 'Prefix to use when naming the output files. defaults to the dataset filename.',
		action = 'store',
		type = str,
		dest = 'name',
		default = None
	)
	##############################################################################################################################################
	# --------------------------------------------------------- Input Data Options ---------------------------------------------------------------
	##############################################################################################################################################
	parser.add_argument(
		"--sheetname",
		help = "Indicates the sheet to use if the input table is an excel workbook and the data is not in Sheet1",
		action = 'store',
		dest = 'sheetname',
		default = 'Sheet1',
		type = str
	)
	parser.add_argument(
		"--genotypes", "--cohorts",
		help = "Indicates that the input table contains genotypes rather than trajectories.",
		action = 'store_true',
		dest = 'is_genotype'
	)

	##############################################################################################################################################
	# ------------------------------------------------------ General Analysis Options ------------------------------------------------------------
	##############################################################################################################################################
	parser.add_argument(
		'--fixed',
		help = "The minimum frequency at which to consider a mutation fixed.",
		action = FixedBreakpointParser,
		dest = 'fixed_breakpoint',
		type = float
	)
	parser.add_argument(
		"-d", "--detection",
		help = "The uncertainty to apply when performing frequency-based calculations. \
			For example, a frequency at a given timepoint is considered undetected if it falls below 0 + `uncertainty`.",
		action = 'store',
		default = 0.03,
		dest = 'detection_breakpoint',
		type = float
	)
	parser.add_argument(
		"-s", "--significant",
		help = "The frequency at which to consider a genotype significantly greater than zero.",
		action = 'store',
		default = 0.15,
		dest = "significant_breakpoint",
		type = float
	)
	parser.add_argument(
		"--additive",
		help = "Controls how the additive score between a nested and unnested genotype is calculated. Defaults to the detection cutoff value.",
		action = 'store',
		default = None,
		dest = 'additive_cutoff',
		type = float
	)
	parser.add_argument(
		"--subtractive",
		help = "Controls when the combined frequencies of a nested and unnested genotype are considered consistently larger than the fixed cutoff."
			   "Defaults to the detection cutoff value.",
		default = None,
		dest = "subtractive_cutoff"
	)
	parser.add_argument(
		"--derivative",
		help = "Controls how much a nested and unnested genotype should be correlated/anticorrelated to be considered significant",
		default = 0.01,
		dest = "derivative_cutoff",
		type = float
	)

	##############################################################################################################################################
	# ----------------------------------------------- Options for individual analysis steps ------------------------------------------------------
	##############################################################################################################################################

	parser.add_argument(
		"-f", "--frequencies",
		help = 'The frequency cutoff to use when sorting the muller_genotypes by first detected frequency. For example, a value of 0.15 will use the frequencies 0,.15,.30,.45...',
		action = FrequencyParser,
		dest = 'frequencies',
		default = [1.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.0]
	)
	parser.add_argument(
		"--similarity-cutoff",
		help = "Maximum p-value difference to consider trajectories related. Used when grouping trajectories into genotypes.",
		action = "store",
		default = 0.05,
		dest = "similarity_breakpoint",
		type = float
	)
	parser.add_argument(
		"--difference-cutoff",
		help = "Minimum p-value to consider a pair of muller_genotypes unrelated. Used when splitting muller_genotypes.",
		action = "store",
		default = 0.25,
		dest = "difference_breakpoint",
		type = float
	)

	##############################################################################################################################################
	# --------------------------------------------------- Genotype Clustering Options ------------------------------------------------------------
	##############################################################################################################################################
	parser.add_argument(
		"--no-filter",
		help = "Disables genotype filtering.",
		action = 'store_false',
		dest = 'use_filter'
	)
	parser.add_argument(
		"--strict-filter",
		help = """By default, the filters allow trajectories to appear both before and after a genotype"""
			   """fixes as long as they were undetected at the timepoint the sweep occurs. This generally"""
			   """represents mutations which appear, are removed during a genotype sweep, and reappear """
			   """afterwards. Using `--strict-filter` would remove these trajectories.""",
		action = "store_true",
		dest = "use_strict_filter"
	)
	parser.add_argument(
		'-m', '--method',
		help = "The clustering method to use. `matlab` will use the original two-step algorithm while `hierarchy` will use hierarchical clustering.",
		action = "store",
		default = "hierarchy",
		dest = "method",
		choices = ['matlab', 'hierarchy', 'twostep']
	)
	parser.add_argument(
		"--metric",
		help = "Selects the distance metric to use. Each metric tends to focus on a specific feature between two series, " \
			   "such as the difference between them or how well they are correlated.",
		action = "store",
		default = "binomial",
		dest = "metric",
		choices = ['similarity', 'binomial', 'pearson', 'minkowski', 'jaccard', 'combined']
	)
	##############################################################################################################################################
	# -------------------------------------------------------- Graphics Options ------------------------------------------------------------------
	##############################################################################################################################################
	parser.add_argument(
		"--annotate-all",
		help = "Adds all gene labels to the muller plots, instead of the top three.",
		action = "store_true",
		dest = "annotate_all"
	)
	parser.add_argument(
		"--genotype-colors",
		help = "An optional map of genotypes to specified colors.",
		action = "store",
		type = Path,
		default = None,
		dest = "genotype_palette_filename"
	)

	##############################################################################################################################################
	# ----------------------------------------------------- Additional Input Files ---------------------------------------------------------------
	##############################################################################################################################################
	parser.add_argument(
		"--gene-alias",
		help = "An optional two-column file with more accurate gene names. This is usefull when using a reference annotated via prokka.",
		action = "store",
		type = Path,
		default = None,
		dest = 'alias_filename'
	)
	parser.add_argument(
		"-g", "--known-genotypes",
		help = "A file with trajectories known to be in the same genotypes. "
			   "Each genotype is defined by a comma-delimited line with the labels of the member trajectories.",
		action = "store",
		default = None,
		dest = "known_genotypes"
	)
	parser.add_argument(
		"--known-ancestry",
		help = "A file designating the known ancestry of certain genotypes. Formatted like the ggmuller edges table.",
		dest = 'known_ancestry',
		default = None,
		type = Path
	)


	##############################################################################################################################################
	# ------------------------------------------------------- Graphics Options -------------------------------------------------------------------
	##############################################################################################################################################

	parser.add_argument(
		"--no-outline",
		help = 'Disables the white outline in the muller plots.',
		action = 'store_true'
	)
	return parser



if __name__ == "__main__":
	import sys
	from pathlib import Path
	sys.path.append(str(Path(__file__).parent.parent))
	from muller.commandline_parser import create_parser
	from muller.muller_workflow import workflow
	args = create_parser().parse_args()

	workflow(args.filename, args.output_folder, program_options = args)


from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
from scipy.cluster import hierarchy

try:
	from muller.clustering.metrics import PairwiseCalculationCache
except ModuleNotFoundError:
	from ..clustering.metrics import PairwiseCalculationCache


def plot_dendrogram(linkage_table: Any, pair_array: PairwiseCalculationCache, filename: Path):
	fig, ax = plt.subplots(figsize = (15, 15))
	# plt.figure(figsize = (15, 15))
	ax.set_title('Hierarchical Clustering Dendrogram', size = 30)
	ax.set_xlabel('Trajectory Label', size = 20)
	ax.set_ylabel('Distance', size = 20)
	hierarchy.dendrogram(
		linkage_table,
		leaf_rotation = 90,  # rotates the x axis labels
		leaf_font_size = 8,  # font size for the x axis labels,
		labels = pair_array.squareform().index,
		ax = ax
	)

	plt.savefig(filename, dpi = 500)
